import cv2
import sys
import serial
from collections import namedtuple
import time


# Constants
SENSITIVITY_VALUE = 40
BLUR_SIZE = 10

# Detection bounding box
theObject = namedtuple("Point", "x y")
objectBoundingRect = namedtuple("Rect", "x width y height")
ser = serial.Serial('/dev/ttyUSB0', 9600)


def search_for_movement(threshold_image, camera_feed):
    global theObject
    global objectBoundingRect

    buf = []

    width = len(thresholdImage[0])
    height = len(thresholdImage)

    center_x = width / 2
    center_y = height / 2

    temp = threshold_image

    temp, contours, hierarchy = cv2.findContours(temp, cv2.RETR_EXTERNAL,
                                                 cv2.CHAIN_APPROX_SIMPLE)

    if len(contours) > 0:
        object_detected = True
    else:
        object_detected = False

    if object_detected:
        largest_contour_vec = [contours[len(contours) - 1]]

        del objectBoundingRect
        objectBoundingRect = cv2.boundingRect(largest_contour_vec[0])

        x_pos = objectBoundingRect[0] + objectBoundingRect[2] / 2
        y_pos = objectBoundingRect[1] + objectBoundingRect[3] / 2

        data = list(str(abs(x_pos - center_x)))

        # Begin transmission
        ser.write(bytearray([0]))
        ser.read()

        # Send sign
        if x_pos - center_x < 0:
            ser.write(bytearray(['-']))
        else:
            ser.write(bytearray(['+']))

        # Send number
        for i in range(0, len(data)):
            ser.write(bytearray([data[i]]))
            buf.append(ser.read())

        # End transmission
        ser.write(bytearray([13]))
        ser.read()

        del theObject
        theObject = namedtuple("Point", "x y")
        theObject.x = x_pos
        theObject.y = y_pos

        x = theObject.x
        y = theObject.y

        cv2.circle(camera_feed, (x, y), 20, (255, 0, 0), 2)
        cv2.line(camera_feed, (x, y), (x, y - 25), (255, 0, 0), 2)
        cv2.line(camera_feed, (x, y), (x, y + 25), (255, 0, 0), 2)
        cv2.line(camera_feed, (x, y), (x - 25, y), (255, 0, 0), 2)
        cv2.line(camera_feed, (x, y), (x + 25, y), (255, 0, 0), 2)
        cv2.putText(camera_feed, "Tracking object at (" +
                    str(x) + ", " + str(y) + ")",
                    (x, y), 1, 1, (255, 0, 0), 2)


if __name__ == '__main__':
    objectDetected = False
    debugMode = False
    trackingEnabled = False
    pause = False
    debugWindowOpen = False

    frame1 = None
    frame2 = None
    grayImage1 = None
    grayImage2 = None
    differenceImage = None
    thresholdImage = None

    cap = cv2.VideoCapture(0)

    if not cap.isOpened():
        print("Could not access camera")
        sys.exit()

    while True:
        # read first frame
        ok, frame1 = cap.read()

        if not ok:
            print("No value returned")
            sys.exit()

        # convert frame1 to gray scale for frame differencing
        grayImage1 = cv2.cvtColor(frame1, cv2.COLOR_BGR2GRAY)

        # copy second frame
        ok, frame2 = cap.read()

        # convert frame2 to gray scale for frame differencing
        grayImage2 = cv2.cvtColor(frame2, cv2.COLOR_BGR2GRAY)

        # perform frame differencing with the sequential images.
        # This will output an "intensity image"
        differenceImage = cv2.absdiff(grayImage1, grayImage2)

        if differenceImage is None:
            print("Error absdiff")
            sys.exit()

        # threshold intensity image at a given sensitivity value
        ok, thresholdImage = cv2.threshold(differenceImage, SENSITIVITY_VALUE,
                                           255, cv2.THRESH_BINARY)

        if not ok:
            print("Error threshold")
            sys.exit()

        if debugMode:
            # show the difference image and threshold image
            cv2.imshow("Difference image", differenceImage)
            cv2.imshow("Threshold image", thresholdImage)
            debugWindowOpen = True
        elif debugWindowOpen:
            # if not in debug mode, destroy the windows
            # so we don't see them anymore
            cv2.destroyWindow("Difference image")
            cv2.destroyWindow("Threshold image")

        # use blur() to smooth the image, remove possible noise and
        # increase the size of the object we are trying to track.
        thresholdImage = cv2.blur(thresholdImage, (BLUR_SIZE, BLUR_SIZE))

        # threshold again to obtain binary image from blur output
        ok, thresholdImage = cv2.threshold(thresholdImage, SENSITIVITY_VALUE,
                                           255, cv2.THRESH_BINARY)

        if not ok:
            print("Error threshold blur")
            sys.exit()

        if debugMode:
            # show the threshold image after it's been "blurred"
            cv2.imshow("Blurred threshold image", thresholdImage)
        elif debugWindowOpen:
            # if not in debug mode, destroy the windows
            # so we don't see them anymore
            cv2.destroyWindow("Blurred threshold image")
            debugWindowOpen = False

        # if tracking enabled, search for contours in our thresholded image
        search_for_movement(thresholdImage, frame1)

        # show our captured frame
        cv2.imshow("Frame 1", frame1)

        # check to see if a button has been pressed. this 10ms delay is
        # necessary for proper operation of this program if removed,
        # frames will not have enough time to referesh and a blank image
        # will appear.
        key = cv2.waitKey(10) & 0xff

        if key == 27:
            # 'esc' key has been pressed, exit program.
            break
        elif key == 116:
            # 't' has been pressed. this will toggle tracking
            trackingEnabled = not trackingEnabled

            if not trackingEnabled:
                print("Tracking disabled")
            else:
                print("Tracking enabled")
        elif key == 100:
            # 'd' has been pressed. this will debug mode
            debugMode = not debugMode

            if not debugMode:
                print("Debug mode disabled")
            else:
                print("Debug mode enabled")
        elif key == 112:
            # 'p' has been pressed. this will pause/resume the code.
            pause = not pause

            if pause:
                print("Code paused, press 'p' again to resume")

                while pause:
                    # Wait for the user to press 'p' again
                    key2 = cv2.waitKey()

                    if key2 == 112:
                        pause = False
                        print("Code resumed")
                        break

    cap.release()
    cv2.destroyAllWindows()
